from datetime import date
from io import BytesIO
from uuid import UUID, uuid4

import pytest
from clowmdb.models import Workflow, WorkflowExecution, WorkflowMode, WorkflowVersion
from fastapi import status
from httpx import AsyncClient
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession

from app.core.config import settings
from app.schemas.workflow import WorkflowIn, WorkflowOut, WorkflowUpdate
from app.schemas.workflow_mode import WorkflowModeIn
from app.scm import SCM
from app.tests.mocks.mock_s3_resource import MockS3ServiceResource
from app.tests.utils.user import UserWithAuthHeader
from app.tests.utils.utils import random_hex_string, random_lower_string


class _TestWorkflowRoutes:
    base_path: str = "/workflows"


class TestWorkflowRoutesCreate(_TestWorkflowRoutes):
    @pytest.mark.asyncio
    async def test_create_workflow_with_github(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Exhaustive Test for successfully creating a workflow.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        git_commit_hash = random_hex_string()
        workflow = WorkflowIn(
            git_commit_hash=git_commit_hash,
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://github.de/example-user/example",
        ).model_dump()
        response = await client.post(self.base_path, json=workflow, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_201_CREATED
        created_workflow = response.json()
        assert not created_workflow["private"]

        stmt = select(Workflow).where(Workflow._workflow_id == UUID(hex=created_workflow["workflow_id"]).bytes)
        db_workflow = await db.scalar(stmt)
        assert db_workflow is not None

        stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == git_commit_hash)
        db_version = await db.scalar(stmt)
        assert db_version is not None

        await db.execute(
            delete(Workflow).where(Workflow._workflow_id == UUID(hex=created_workflow["workflow_id"]).bytes)
        )
        await db.commit()

    @pytest.mark.asyncio
    async def test_create_workflow_with_gitlab(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for successfully creating a workflow with a Gitlab repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        workflow = WorkflowIn(
            git_commit_hash=random_hex_string(),
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://gitlab.de/example-user/example",
        ).model_dump()
        response = await client.post(self.base_path, json=workflow, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_201_CREATED
        created_workflow = response.json()
        assert not created_workflow["private"]

        stmt = select(Workflow).where(Workflow._workflow_id == UUID(hex=created_workflow["workflow_id"]).bytes)
        db_workflow = await db.scalar(stmt)
        assert db_workflow is not None

        # Download SCM file in PARAMS_BUCKET that should be created
        scm_file_name = f"{db_workflow.workflow_id.hex}.scm"
        assert scm_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        obj = mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(scm_file_name)
        with BytesIO() as f:
            obj.download_fileobj(f)
            f.seek(0)
            scm = SCM.deserialize(f)

        # Check content of SCM file
        assert len(scm.providers) == 1
        provider = scm.providers[0]
        assert provider.password is None
        assert provider.name == f"repo{db_workflow.workflow_id.hex}"
        assert provider.platform == "gitlab"
        assert provider.server == "https://gitlab.de"

        await db.execute(
            delete(Workflow).where(Workflow._workflow_id == UUID(hex=created_workflow["workflow_id"]).bytes)
        )
        await db.commit()
        obj.delete()

    @pytest.mark.asyncio
    async def test_create_workflow_with_private_gitlab(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for successfully creating a workflow with a private Gitlab repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        token = random_lower_string(20)
        workflow = WorkflowIn(
            git_commit_hash=random_hex_string(),
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://gitlab.de/example-user/example",
            token=token,
        ).model_dump()
        response = await client.post(self.base_path, json=workflow, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_201_CREATED
        created_workflow = response.json()
        assert created_workflow["private"]

        # Check if workflow is created in database
        stmt = select(Workflow).where(Workflow._workflow_id == UUID(hex=created_workflow["workflow_id"]).bytes)
        db_workflow = await db.scalar(stmt)
        assert db_workflow is not None
        # Check if token is saved
        assert db_workflow.credentials_token == token

        # Download SCM file in PARAMS_BUCKET that should be created
        scm_file_name = f"{db_workflow.workflow_id.hex}.scm"
        assert scm_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        obj = mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(scm_file_name)
        with BytesIO() as f:
            obj.download_fileobj(f)
            f.seek(0)
            scm = SCM.deserialize(f)

        # Check content of SCM file
        assert len(scm.providers) == 1
        provider = scm.providers[0]
        assert provider.password == token
        assert provider.name == f"repo{db_workflow.workflow_id.hex}"
        assert provider.platform == "gitlab"
        assert provider.server == "https://gitlab.de"

        # Cleanup after test
        await db.execute(delete(Workflow).where(Workflow._workflow_id == db_workflow.workflow_id.bytes))
        await db.commit()
        obj.delete()

    @pytest.mark.asyncio
    async def test_create_workflow_with_private_github(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for successfully creating a workflow with a private GitHub repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        token = random_lower_string(20)
        workflow = WorkflowIn(
            git_commit_hash=random_hex_string(),
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://github.de/example-user/example",
            token=token,
        ).model_dump()
        response = await client.post(self.base_path, json=workflow, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_201_CREATED
        created_workflow = response.json()
        assert created_workflow["private"]

        # Check if workflow is created in database
        stmt = select(Workflow).where(Workflow._workflow_id == UUID(hex=created_workflow["workflow_id"]).bytes)
        db_workflow = await db.scalar(stmt)
        assert db_workflow is not None
        # Check if token is saved
        assert db_workflow.credentials_token == token

        # Download SCM file in PARAMS_BUCKET that should be created
        scm_file_name = f"{db_workflow.workflow_id.hex}.scm"
        assert scm_file_name in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()
        obj = mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(scm_file_name)
        with BytesIO() as f:
            obj.download_fileobj(f)
            f.seek(0)
            scm = SCM.deserialize(f)

        # Check content of SCM file
        assert len(scm.providers) == 1
        provider = scm.providers[0]
        assert provider.password == token
        assert provider.name == "github"
        assert provider.user == "example-user"

        # Cleanup after test
        await db.execute(delete(Workflow).where(Workflow._workflow_id == db_workflow.workflow_id.bytes))
        await db.commit()
        obj.delete()

    @pytest.mark.asyncio
    async def test_create_workflow_with_error(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for creating a workflow where the file checks don't pass

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        workflow = WorkflowIn(
            git_commit_hash=random_hex_string(),
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://github.de/example-user/example",
        ).model_dump()
        response = await client.post(
            self.base_path, params={"raise_error": True}, json=workflow, headers=random_user.auth_headers
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST

        db_workflows = (await db.scalars(select(Workflow))).all()
        assert len(db_workflows) == 0

    @pytest.mark.asyncio
    async def test_create_workflow_with_taken_name(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for creating a workflow with an already taken name.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow = WorkflowIn(
            git_commit_hash=random_hex_string(),
            name=random_workflow.name,
            short_description=random_lower_string(65),
            repository_url="https://github.de/example-user/example",
        ).model_dump()
        response = await client.post(self.base_path, json=workflow, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_create_workflow_with_registered_git_commit(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for creating a workflow where the git commit is already in the system.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        workflow = WorkflowIn(
            git_commit_hash=random_workflow.versions[0].git_commit_hash,
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://github.de/example-user/example",
        ).model_dump()
        response = await client.post(self.base_path, json=workflow, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_create_workflow_with_unknown_git_repository(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for creating a workflow with an unknown git repository.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        workflow = WorkflowIn(
            git_commit_hash=random_hex_string(),
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://example.org",
        ).model_dump()
        response = await client.post(self.base_path, json=workflow, headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_create_workflow_with_workflow_mode(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Exhaustive Test for successfully creating a workflow with a workflow mode.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        git_commit_hash = random_hex_string()
        workflow_mode = WorkflowModeIn(
            name=random_lower_string(), entrypoint=random_lower_string(), schema_path=random_lower_string()
        )
        workflow = WorkflowIn(
            git_commit_hash=git_commit_hash,
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://github.de/example-user/example",
            modes=[workflow_mode, workflow_mode],
        )
        response = await client.post(self.base_path, json=workflow.model_dump(), headers=random_user.auth_headers)
        assert response.status_code == status.HTTP_201_CREATED
        created_workflow = response.json()
        assert len(created_workflow["versions"][0]["modes"]) == 2
        mode_id = UUID(hex=created_workflow["versions"][0]["modes"][0])

        stmt = select(WorkflowMode).where(WorkflowMode._mode_id == mode_id.bytes)
        db_mode = await db.scalar(stmt)
        assert db_mode is not None
        assert db_mode.name == workflow_mode.name
        assert db_mode.entrypoint == workflow_mode.entrypoint
        assert db_mode.schema_path == workflow_mode.schema_path

        assert (
            mock_s3_service.Bucket(settings.WORKFLOW_BUCKET)  # type: ignore[func-returns-value]
            .Object(f"{workflow.git_commit_hash}-{db_mode.mode_id.hex}.json")
            .load()
            is None
        )

        # Clean up after test
        await db.execute(
            delete(Workflow).where(Workflow._workflow_id == UUID(hex=created_workflow["workflow_id"]).bytes)
        )
        await db.execute(delete(WorkflowMode).where(WorkflowMode._mode_id == mode_id.bytes))
        await db.commit()


class TestWorkflowRoutesList(_TestWorkflowRoutes):
    @pytest.mark.asyncio
    async def test_list_all_unpublished_workflows(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for listing all workflows in the system.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.get(
            self.base_path,
            headers=random_user.auth_headers,
            params={"version_status": WorkflowVersion.Status.CREATED.name},
        )
        assert response.status_code == status.HTTP_200_OK
        workflows = response.json()
        assert len(workflows) == 1
        assert workflows[0]["workflow_id"] == str(random_workflow.workflow_id)
        assert not workflows[0]["private"]

    @pytest.mark.asyncio
    async def test_list_all_workflows(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for listing all workflows in the system.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.get(
            self.base_path,
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        workflows = response.json()
        assert len(workflows) == 0

    @pytest.mark.asyncio
    async def test_list_workflows_by_developer(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        random_second_user: UserWithAuthHeader,
    ) -> None:
        """
        Test for listing all workflows in the system.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.get(
            self.base_path,
            headers=random_second_user.auth_headers,
            params={"developer_id": random_user.user.uid, "version_status": WorkflowVersion.Status.CREATED.name},
        )
        assert response.status_code == status.HTTP_200_OK
        workflows = response.json()
        assert len(workflows) == 1
        assert workflows[0]["workflow_id"] == str(random_workflow.workflow_id)


class TestWorkflowRoutesGet(_TestWorkflowRoutes):
    @pytest.mark.asyncio
    async def test_get_workflow_with_version_status(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for getting a workflow by its id.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        response = await client.get(
            "/".join([self.base_path, str(random_workflow.workflow_id)]),
            params={"version_status": random_workflow.versions[0].status.name},
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        workflow = response.json()
        assert workflow["workflow_id"] == str(random_workflow.workflow_id)
        assert not workflow["private"]
        assert len(workflow["versions"]) > 0

    @pytest.mark.asyncio
    async def test_get_non_existing_workflow(self, client: AsyncClient, random_user: UserWithAuthHeader) -> None:
        """
        Test for getting a non-existing workflow.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        response = await client.get(
            "/".join([self.base_path, str(uuid4())]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    @pytest.mark.asyncio
    async def test_get_workflow_statistics(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        random_workflow_execution: WorkflowExecution,
    ) -> None:
        """
        Test for getting a non-existing workflow.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        response = await client.get(
            "/".join([self.base_path, str(random_workflow.workflow_id), "statistics"]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_200_OK
        statistics = response.json()
        assert len(statistics) == 1
        assert statistics[0]["day"] == str(date.today())
        assert statistics[0]["count"] == 1


class TestWorkflowRoutesDelete(_TestWorkflowRoutes):
    @pytest.mark.asyncio
    async def test_delete_workflow(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for deleting a workflow by its id.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        response = await client.delete(
            "/".join([self.base_path, str(random_workflow.workflow_id)]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT
        icon_slug = str(random_workflow.versions[0].icon_url).split("/")[-1]
        assert icon_slug not in mock_s3_service.Bucket(settings.ICON_BUCKET).objects.all_keys()
        schema_file = random_workflow.versions[0].git_commit_hash + ".json"
        assert schema_file not in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()

    @pytest.mark.asyncio
    async def test_delete_private_workflow(
        self,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_private_workflow: WorkflowOut,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for deleting a workflow by its id.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_private_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        response = await client.delete(
            "/".join([self.base_path, str(random_private_workflow.workflow_id)]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT
        icon_slug = str(random_private_workflow.versions[0].icon_url).split("/")[-1]
        assert icon_slug not in mock_s3_service.Bucket(settings.ICON_BUCKET).objects.all_keys()
        schema_file = random_private_workflow.versions[0].git_commit_hash + ".json"
        assert schema_file not in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()
        scm_file = f"{random_private_workflow.workflow_id.hex}.scm"
        assert scm_file not in mock_s3_service.Bucket(settings.PARAMS_BUCKET).objects.all_keys()

    @pytest.mark.asyncio
    async def test_delete_workflow_with_mode(
        self,
        client: AsyncClient,
        db: AsyncSession,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        random_workflow_mode: WorkflowMode,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for deleting a workflow by its id.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        random_workflow_mode : clowmdb.model.WorkflowMode
            Random workflow mode for testing
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        """
        response = await client.delete(
            "/".join([self.base_path, str(random_workflow.workflow_id)]),
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT
        icon_slug = str(random_workflow.versions[0].icon_url).split("/")[-1]
        assert icon_slug not in mock_s3_service.Bucket(settings.ICON_BUCKET).objects.all_keys()
        schema_file = f"{random_workflow.versions[0].git_commit_hash}-{random_workflow_mode.mode_id.hex}.json"
        assert schema_file not in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()

        mode_db = await db.scalar(
            select(WorkflowMode).where(WorkflowMode._mode_id == random_workflow_mode.mode_id.bytes)
        )
        assert mode_db is None


class TestWorkflowRoutesUpdate(_TestWorkflowRoutes):
    @pytest.mark.asyncio
    async def test_update_workflow(
        self, db: AsyncSession, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for successfully updating a workflow.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        git_commit_hash = random_hex_string()
        version_update = WorkflowUpdate(
            git_commit_hash=git_commit_hash,
            version=random_lower_string(8),
        ).model_dump()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            json=version_update,
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_201_CREATED
        created_version = response.json()
        assert created_version["git_commit_hash"] == git_commit_hash
        assert created_version["status"] == WorkflowVersion.Status.CREATED
        assert created_version["icon_url"] == str(random_workflow.versions[0].icon_url)

        stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == git_commit_hash)
        db_version = await db.scalar(stmt)
        assert db_version is not None
        assert db_version.status == WorkflowVersion.Status.CREATED

    @pytest.mark.asyncio
    async def test_update_workflow_with_append_modes(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        random_workflow_mode: WorkflowMode,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for successfully updating a workflow and adding new modes.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        random_workflow_mode : clowmdb.model.WorkflowMode
            Random workflow mode for testing
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        git_commit_hash = random_hex_string()
        version_update = WorkflowUpdate(
            git_commit_hash=git_commit_hash,
            version=random_lower_string(8),
            append_modes=[
                WorkflowModeIn(
                    name=random_lower_string(10), entrypoint=random_lower_string(16), schema_path=random_lower_string()
                )
            ],
        ).model_dump()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            json=version_update,
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_201_CREATED
        created_version = response.json()
        assert created_version["git_commit_hash"] == git_commit_hash
        assert created_version["status"] == WorkflowVersion.Status.CREATED
        assert created_version["icon_url"] == str(random_workflow.versions[0].icon_url)
        assert created_version["modes"] is not None
        assert len(created_version["modes"]) == 2

        stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == git_commit_hash)
        db_version = await db.scalar(stmt)
        assert db_version is not None
        assert db_version.status == WorkflowVersion.Status.CREATED

        new_mode_id = next((UUID(m) for m in created_version["modes"] if m != str(random_workflow_mode.mode_id)), None)
        assert new_mode_id is not None
        assert (
            f"{git_commit_hash}-{new_mode_id.hex}.json"
            in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()
        )
        assert (
            f"{git_commit_hash}-{random_workflow_mode.mode_id.hex}.json"
            in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()
        )

        # Clean up after test
        await db.execute(delete(WorkflowMode).where(WorkflowMode._mode_id == new_mode_id.bytes))
        await db.commit()
        mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(f"{git_commit_hash}-{new_mode_id.hex}.json").delete()

    @pytest.mark.asyncio
    async def test_update_workflow_with_delete_non_existing_modes(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
    ) -> None:
        """
        Test for updating a workflow and delete an non-existing mode.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        git_commit_hash = random_hex_string()
        version_update = WorkflowUpdate(
            git_commit_hash=git_commit_hash, version=random_lower_string(8), delete_modes=[str(uuid4())]
        ).model_dump_json()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            data=version_update,  # type: ignore[arg-type]
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_update_workflow_with_delete_modes(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        random_workflow_mode: WorkflowMode,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for successfully updating a workflow and delete an old mode.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        random_workflow_mode : clowmdb.model.WorkflowMode
            Random workflow mode for testing
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        git_commit_hash = random_hex_string()
        version_update = WorkflowUpdate(
            git_commit_hash=git_commit_hash,
            version=random_lower_string(8),
            delete_modes=[str(random_workflow_mode.mode_id)],
        ).model_dump_json()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            data=version_update,  # type: ignore[arg-type]
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_201_CREATED
        created_version = response.json()
        assert created_version["git_commit_hash"] == git_commit_hash
        assert created_version["status"] == WorkflowVersion.Status.CREATED
        assert created_version["icon_url"] == str(random_workflow.versions[0].icon_url)
        assert created_version["modes"] is not None
        assert created_version["modes"] is None or len(created_version["modes"]) == 0

        stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == git_commit_hash)
        db_version = await db.scalar(stmt)
        assert db_version is not None
        assert db_version.status == WorkflowVersion.Status.CREATED

        assert (
            f"{git_commit_hash}-{random_workflow_mode.mode_id.hex}.json"
            not in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()
        )

    @pytest.mark.asyncio
    async def test_update_workflow_with_append_and_delete_modes(
        self,
        db: AsyncSession,
        client: AsyncClient,
        random_user: UserWithAuthHeader,
        random_workflow: WorkflowOut,
        random_workflow_mode: WorkflowMode,
        mock_s3_service: MockS3ServiceResource,
    ) -> None:
        """
        Test for successfully updating a workflow with adding a new mode and delete an old mode.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        random_workflow_mode : clowmdb.model.WorkflowMode
            Random workflow mode for testing
        mock_s3_service : app.tests.mocks.mock_s3_resource.MockS3ServiceResource
            Mock S3 Service to manipulate objects.
        """
        git_commit_hash = random_hex_string()
        version_update = WorkflowUpdate(
            git_commit_hash=git_commit_hash,
            version=random_lower_string(8),
            delete_modes=[str(random_workflow_mode.mode_id)],
            append_modes=[
                WorkflowModeIn(
                    name=random_lower_string(10), entrypoint=random_lower_string(16), schema_path=random_lower_string()
                )
            ],
        ).model_dump_json()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            data=version_update,  # type: ignore[arg-type]
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_201_CREATED
        created_version = response.json()
        assert created_version["git_commit_hash"] == git_commit_hash
        assert created_version["status"] == WorkflowVersion.Status.CREATED
        assert created_version["icon_url"] == str(random_workflow.versions[0].icon_url)
        assert len(created_version["modes"]) == 1
        assert created_version["modes"][0] != str(random_workflow_mode.mode_id)

        stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == git_commit_hash)
        db_version = await db.scalar(stmt)
        assert db_version is not None
        assert db_version.status == WorkflowVersion.Status.CREATED

        mode_id = UUID(created_version["modes"][0])
        assert (
            f"{git_commit_hash}-{mode_id.hex}.json"
            in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()
        )
        assert (
            f"{git_commit_hash}-{random_workflow_mode.mode_id.hex}.json"
            not in mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).objects.all_keys()
        )

        # Clean up after test
        await db.execute(delete(WorkflowMode).where(WorkflowMode._mode_id == mode_id.bytes))
        await db.commit()
        mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(f"{git_commit_hash}-{mode_id.hex}.json").delete()

    @pytest.mark.asyncio
    async def test_update_workflow_with_error(
        self, db: AsyncSession, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for updating a workflow where the file checks don't pass

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        version_update = WorkflowUpdate(
            git_commit_hash=random_hex_string(),
            version=random_lower_string(8),
        ).model_dump()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            params={"raise_error": True},
            json=version_update,
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST

        db_workflows = (
            await db.scalars(
                select(WorkflowVersion).where(WorkflowVersion._workflow_id == random_workflow.workflow_id.bytes)
            )
        ).all()
        assert len(db_workflows) == 1

    @pytest.mark.asyncio
    async def test_update_workflow_with_registered_git_commit(
        self, client: AsyncClient, random_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for updating a workflow where the git commit is already in the system.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        version_update = WorkflowUpdate(
            git_commit_hash=random_workflow.versions[0].git_commit_hash,
            version=random_lower_string(8),
        ).model_dump()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            json=version_update,
            headers=random_user.auth_headers,
        )
        assert response.status_code == status.HTTP_400_BAD_REQUEST

    @pytest.mark.asyncio
    async def test_update_workflow_as_foreign_user(
        self, client: AsyncClient, random_second_user: UserWithAuthHeader, random_workflow: WorkflowOut
    ) -> None:
        """
        Test for updating a workflow where the requested user is not the developer of the workflow.

        Parameters
        ----------
        client : httpx.AsyncClient
            HTTP Client to perform the request on.
        random_second_user : app.tests.utils.user.UserWithAuthHeader
            Random second user for testing.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        version_update = WorkflowUpdate(
            git_commit_hash=random_hex_string(),
            version=random_lower_string(8),
        ).model_dump()
        response = await client.post(
            "/".join([self.base_path, str(random_workflow.workflow_id), "update"]),
            json=version_update,
            headers=random_second_user.auth_headers,
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN
