import asyncio
from functools import partial
from io import BytesIO
from secrets import token_urlsafe
from typing import AsyncIterator, Iterator

import httpx
import pytest
import pytest_asyncio
from clowmdb.db.session import get_async_session
from clowmdb.models import (
    Bucket,
    Workflow,
    WorkflowExecution,
    WorkflowMode,
    WorkflowVersion,
    workflow_mode_association_table,
)
from fastapi import status
from sqlalchemy import insert, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from app.api.dependencies import get_db, get_decode_jwt_function, get_httpx_client, get_s3_resource
from app.core.config import settings
from app.git_repository import build_repository
from app.main import app
from app.schemas.workflow import WorkflowOut
from app.scm import SCM, Provider
from app.tests.mocks import MockOpaService, MockS3ServiceResource, MockSlurmCluster
from app.tests.utils.bucket import create_random_bucket
from app.tests.utils.user import UserWithAuthHeader, create_random_user, decode_mock_token, get_authorization_headers
from app.tests.utils.utils import random_hex_string, random_lower_string

jwt_secret = token_urlsafe(32)


@pytest.fixture(scope="session")
def event_loop() -> Iterator:
    """
    Creates an instance of the default event loop for the test session.
    """
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def mock_s3_service() -> Iterator[MockS3ServiceResource]:
    """
    Fixture for creating a mock object for the rgwadmin package.
    """
    mock_s3 = MockS3ServiceResource()
    mock_s3.Bucket(settings.ICON_BUCKET).create()
    mock_s3.Bucket(settings.PARAMS_BUCKET).create()
    mock_s3.Bucket(settings.WORKFLOW_BUCKET).create()
    yield mock_s3
    mock_s3.delete_bucket(name=settings.ICON_BUCKET, force_delete=True)
    mock_s3.delete_bucket(name=settings.PARAMS_BUCKET, force_delete=True)
    mock_s3.delete_bucket(name=settings.WORKFLOW_BUCKET, force_delete=True)


@pytest.fixture(scope="session")
def mock_slurm_cluster() -> Iterator[MockSlurmCluster]:
    mock_slurm = MockSlurmCluster()
    yield mock_slurm
    mock_slurm.reset()


@pytest.fixture(scope="session")
def mock_opa_service() -> Iterator[MockOpaService]:
    mock_opa = MockOpaService()
    yield mock_opa
    mock_opa.reset()


@pytest_asyncio.fixture(scope="module")
async def client(
    mock_s3_service: MockS3ServiceResource,
    db: AsyncSession,
    mock_opa_service: MockOpaService,
    mock_slurm_cluster: MockSlurmCluster,
) -> AsyncIterator[httpx.AsyncClient]:
    """
    Fixture for creating a TestClient and perform HTTP Request on it.
    Overrides serveral dependencies.
    """

    async def get_mock_httpx_client(raise_error: bool = False) -> AsyncIterator[httpx.AsyncClient]:
        # raises an 404 error if the query parameter 'raise_error' is true
        def mock_request_handler(request: httpx.Request) -> httpx.Response:
            url = str(request.url)
            if url.startswith(str(settings.OPA_URI)):
                return mock_opa_service.handle_request(request)
            elif url.startswith(str(settings.SLURM_ENDPOINT)):
                return mock_slurm_cluster.handle_request(request)
            elif raise_error:
                return httpx.Response(status_code=status.HTTP_404_NOT_FOUND, json={})
            return httpx.Response(
                status_code=status.HTTP_200_OK,
                json={
                    # When checking if a file exists in a git repository, the GitHub API expects this in a response
                    "download_url": "https://example.com"
                },
            )

        async with httpx.AsyncClient(transport=httpx.MockTransport(mock_request_handler)) as http_client:
            yield http_client

    app.dependency_overrides[get_httpx_client] = get_mock_httpx_client
    app.dependency_overrides[get_s3_resource] = lambda: mock_s3_service
    app.dependency_overrides[get_decode_jwt_function] = lambda: partial(decode_mock_token, secret=jwt_secret)
    app.dependency_overrides[get_db] = lambda: db
    async with httpx.AsyncClient(app=app, base_url="http://localhost") as ac:
        yield ac
    app.dependency_overrides = {}


@pytest_asyncio.fixture(scope="module")
async def db() -> AsyncIterator[AsyncSession]:
    """
    Fixture for creating a database session to connect to.
    """
    async with get_async_session(
        url=str(settings.SQLALCHEMY_DATABASE_ASYNC_URI), verbose=settings.SQLALCHEMY_VERBOSE_LOGGER
    ) as dbSession:
        yield dbSession


@pytest_asyncio.fixture(scope="function")
async def random_user(db: AsyncSession, mock_opa_service: MockOpaService) -> AsyncIterator[UserWithAuthHeader]:
    """
    Create a random user and deletes him afterward.
    """
    user = await create_random_user(db)
    mock_opa_service.add_user(user.uid, privileged=True)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    mock_opa_service.delete_user(user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_second_user(db: AsyncSession, mock_opa_service: MockOpaService) -> AsyncIterator[UserWithAuthHeader]:
    """
    Create a random second user and deletes him afterward.
    """
    user = await create_random_user(db)
    mock_opa_service.add_user(user.uid)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    mock_opa_service.delete_user(user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="module")
async def random_third_user(db: AsyncSession, mock_opa_service: MockOpaService) -> AsyncIterator[UserWithAuthHeader]:
    """
    Create a random third user and deletes him afterward.
    """
    user = await create_random_user(db)
    mock_opa_service.add_user(user.uid)
    yield UserWithAuthHeader(user=user, auth_headers=get_authorization_headers(uid=user.uid, secret=jwt_secret))
    mock_opa_service.delete_user(user.uid)
    await db.delete(user)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_bucket(
    db: AsyncSession, random_user: UserWithAuthHeader, mock_s3_service: MockS3ServiceResource
) -> AsyncIterator[Bucket]:
    """
    Create a random user and deletes him afterwards.
    """
    bucket = await create_random_bucket(db, random_user.user)
    mock_s3_service.Bucket(name=bucket.name).create()
    yield bucket
    mock_s3_service.delete_bucket(name=bucket.name, force_delete=True)
    await db.delete(bucket)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_workflow(
    db: AsyncSession, random_user: UserWithAuthHeader, mock_s3_service: MockS3ServiceResource
) -> AsyncIterator[WorkflowOut]:
    """
    Create a random workflow and deletes it afterwards.
    """
    workflow_db = Workflow(
        name=random_lower_string(10),
        repository_url="https://github.de/example-user/example",
        short_description=random_lower_string(65),
        developer_id=random_user.user.uid,
    )
    db.add(workflow_db)
    await db.commit()
    icon_slug = random_hex_string()
    workflow_version = WorkflowVersion(
        git_commit_hash=random_hex_string(40),
        version="v1.0.0",
        _workflow_id=workflow_db.workflow_id.bytes,
        icon_slug=icon_slug,
    )
    db.add(workflow_version)
    await db.commit()
    mock_s3_service.Bucket(settings.ICON_BUCKET).Object(icon_slug).upload_fileobj(BytesIO(b"{}"))
    mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(f"{workflow_version.git_commit_hash}.json").upload_fileobj(
        BytesIO(b"{}")
    )
    yield WorkflowOut.from_db_workflow(db_workflow=workflow_db, versions=[workflow_version], load_modes=False)
    mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(f"{workflow_version.git_commit_hash}.json").delete()
    mock_s3_service.Bucket(settings.ICON_BUCKET).Object(icon_slug).delete()
    await db.delete(workflow_db)
    await db.commit()


@pytest_asyncio.fixture(scope="function")
async def random_private_workflow(
    db: AsyncSession,
    random_workflow: WorkflowOut,
    random_workflow_version: WorkflowVersion,
    mock_s3_service: MockS3ServiceResource,
) -> AsyncIterator[WorkflowOut]:
    """
    Transform the random workflow into a private workflow.
    """
    # Update credentials in Database
    token = random_lower_string(15)
    await db.execute(
        update(Workflow)
        .where(Workflow._workflow_id == random_workflow.workflow_id.bytes)
        .values(credentials_token=token)
    )
    await db.commit()

    # Upload SCM file to PARAMS_BUCKET
    scm_provider = Provider.from_repo(
        build_repository(random_workflow.repository_url, random_workflow_version.git_commit_hash, token=token),
        name=f"repo{random_workflow.workflow_id.hex}",
    )
    assert scm_provider is not None
    with BytesIO() as f:
        SCM([scm_provider]).serialize(f)
        f.seek(0)
        mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"{random_workflow.workflow_id.hex}.scm").upload_fileobj(
            f
        )
    yield random_workflow
    # Delete SCM in PARAMS_BUCKET after tests
    mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"{random_workflow.workflow_id.hex}.scm").delete()


@pytest_asyncio.fixture(scope="function")
async def random_workflow_version(db: AsyncSession, random_workflow: WorkflowOut) -> WorkflowVersion:
    """
    Create a random workflow version. Will be deleted, when the workflow is deleted.
    """
    stmt = select(WorkflowVersion).where(WorkflowVersion.git_commit_hash == random_workflow.versions[0].git_commit_hash)
    return await db.scalar(stmt)


@pytest_asyncio.fixture(scope="function")
async def random_workflow_execution(
    db: AsyncSession,
    random_workflow_version: WorkflowVersion,
    random_user: UserWithAuthHeader,
    mock_s3_service: MockS3ServiceResource,
    mock_slurm_cluster: MockSlurmCluster,
) -> AsyncIterator[WorkflowExecution]:
    """
    Create a random workflow execution. Will be deleted, when the user is deleted.
    """
    execution = WorkflowExecution(
        user_id=random_user.user.uid,
        workflow_version_id=random_workflow_version.git_commit_hash,
        slurm_job_id=-1,
    )
    db.add(execution)
    await db.commit()
    slurm_job_id = mock_slurm_cluster.add_workflow_execution({"job": {"name": str(execution.execution_id)}})
    await db.execute(
        update(WorkflowExecution)
        .where(WorkflowExecution._execution_id == execution.execution_id.bytes)
        .values(slurm_job_id=slurm_job_id)
    )
    mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"params-{execution.execution_id.hex}.json").upload_fileobj(
        BytesIO(b"{}")
    )
    yield execution
    mock_s3_service.Bucket(settings.PARAMS_BUCKET).Object(f"params-{execution.execution_id.hex}.json").delete()


@pytest_asyncio.fixture(scope="function")
async def random_workflow_mode(
    db: AsyncSession,
    random_workflow_version: WorkflowVersion,
    mock_s3_service: MockS3ServiceResource,
) -> AsyncIterator[WorkflowMode]:
    """
    Create a random workflow execution. Will be deleted, when the user is deleted.
    """
    mode = WorkflowMode(
        name=random_lower_string(8), schema_path=random_lower_string(32), entrypoint=random_lower_string(16)
    )
    db.add(mode)
    await db.commit()
    await db.execute(
        insert(workflow_mode_association_table),
        [
            {
                "workflow_version_commit_hash": random_workflow_version.git_commit_hash,
                "workflow_mode_id": mode.mode_id.bytes,
            }
        ],
    )
    await db.commit()
    key = f"{random_workflow_version.git_commit_hash}-{mode.mode_id.hex}.json"
    parameter_schema = mock_s3_service.Bucket(settings.WORKFLOW_BUCKET).Object(key)
    parameter_schema.upload_fileobj(BytesIO(b"{}"))
    yield mode
    parameter_schema.delete()
    await db.delete(mode)
    await db.commit()
