import random
from datetime import date

import pytest
from clowmdb.models import Workflow, WorkflowExecution, WorkflowVersion
from sqlalchemy import delete, select
from sqlalchemy.ext.asyncio import AsyncSession

from app.crud import CRUDWorkflow
from app.schemas.workflow import WorkflowIn, WorkflowOut
from app.tests.utils.user import UserWithAuthHeader
from app.tests.utils.utils import random_hex_string, random_lower_string


class TestWorkflowCRUDList:
    @pytest.mark.asyncio
    async def test_get_all_workflows(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test get all workflows from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflows = await CRUDWorkflow.list_workflows(db)
        assert len(workflows) == 1
        assert workflows[0].workflow_id == random_workflow.workflow_id

    @pytest.mark.asyncio
    async def test_get_workflows_by_developer(
        self, db: AsyncSession, random_workflow: WorkflowOut, random_user: UserWithAuthHeader
    ) -> None:
        """
        Test get only workflows from the CRUD Repository by specific developer.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        workflows = await CRUDWorkflow.list_workflows(db, developer_id=random_user.user.uid)
        assert len(workflows) == 1
        assert workflows[0].workflow_id == random_workflow.workflow_id

    @pytest.mark.asyncio
    async def test_get_workflows_with_unpublished_version(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test get only workflows from the CRUD Repository with an unpublished version.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflows = await CRUDWorkflow.list_workflows(db, version_status=[WorkflowVersion.Status.CREATED])
        assert len(workflows) == 1
        assert workflows[0].workflow_id == random_workflow.workflow_id

    @pytest.mark.asyncio
    async def test_get_workflow_with_name_substring(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test get workflows with a substring in their name from the CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        substring_indices = [0, 0]
        while substring_indices[0] == substring_indices[1]:
            substring_indices = sorted(random.choices(range(len(random_workflow.name)), k=2))

        random_substring = random_workflow.name[substring_indices[0] : substring_indices[1]]
        workflows = await CRUDWorkflow.list_workflows(db, name_substring=random_substring)
        assert len(workflows) > 0
        assert random_workflow.workflow_id in map(lambda w: w.workflow_id, workflows)

    @pytest.mark.asyncio
    async def test_search_non_existing_workflow_by_name(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for getting a non-existing workflow by its name from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflows = await CRUDWorkflow.list_workflows(db, name_substring=2 * random_workflow.name)
        assert sum(1 for w in workflows if w.workflow_id == random_workflow.workflow_id) == 0


class TestWorkflowCRUDGet:
    @pytest.mark.asyncio
    async def test_get_specific_workflow(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for getting a workflow by its id from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow = await CRUDWorkflow.get(db, random_workflow.workflow_id)
        assert workflow is not None
        assert workflow.workflow_id == random_workflow.workflow_id

    @pytest.mark.asyncio
    async def test_get_workflow_by_name(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for getting a workflow by its name from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        workflow = await CRUDWorkflow.get_by_name(db, random_workflow.name)
        assert workflow is not None
        assert workflow.workflow_id == random_workflow.workflow_id

    @pytest.mark.asyncio
    async def test_get_workflow_statistics(
        self, db: AsyncSession, random_workflow: WorkflowOut, random_workflow_execution: WorkflowExecution
    ) -> None:
        """
        Test for getting a workflow by its name from CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random bucket for testing.
        """
        statistics = await CRUDWorkflow.statistics(db, random_workflow.workflow_id)
        assert len(statistics) == 1
        assert statistics[0].day == date.today()
        assert statistics[0].count == 1


class TestWorkflowCRUDCreate:
    @pytest.mark.asyncio
    async def test_create_workflow(self, db: AsyncSession, random_user: UserWithAuthHeader) -> None:
        """
        Test for creating a workflow in CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_user : app.tests.utils.user.UserWithAuthHeader
            Random user for testing.
        """
        workflow_in = WorkflowIn(
            git_commit_hash=random_hex_string(),
            name=random_lower_string(10),
            short_description=random_lower_string(65),
            repository_url="https://github.com/example/example",
        )
        workflow = await CRUDWorkflow.create(db, workflow=workflow_in, developer=random_user.user.uid)
        assert workflow is not None

        stmt = select(Workflow).where(Workflow._workflow_id == workflow.workflow_id.bytes)
        created_workflow = await db.scalar(stmt)
        assert created_workflow is not None
        assert created_workflow == workflow

        await db.execute(delete(Workflow).where(Workflow._workflow_id == workflow.workflow_id.bytes))
        await db.commit()

    @pytest.mark.asyncio
    async def test_create_workflow_credentials(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for updating the workflow credentials in CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """
        token = random_lower_string(15)
        await CRUDWorkflow.update_credentials(db, workflow_id=random_workflow.workflow_id, token=token)

        stmt = select(Workflow).where(Workflow._workflow_id == random_workflow.workflow_id.bytes)
        workflow = await db.scalar(stmt)
        assert workflow is not None
        assert workflow.credentials_token == token


class TestWorkflowCRUDDelete:
    @pytest.mark.asyncio
    async def test_delete_workflow(self, db: AsyncSession, random_workflow: WorkflowOut) -> None:
        """
        Test for deleting a workflow in CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_workflow : app.schemas.workflow.WorkflowOut
            Random workflow for testing.
        """

        await CRUDWorkflow.delete(db, workflow_id=random_workflow.workflow_id)

        stmt = select(Workflow).where(Workflow._workflow_id == random_workflow.workflow_id.bytes)
        created_workflow = await db.scalar(stmt)
        assert created_workflow is None

        stmt = select(WorkflowVersion).where(WorkflowVersion._workflow_id == random_workflow.workflow_id.bytes)
        versions = (await db.scalars(stmt)).all()

        assert len(versions) == 0

    @pytest.mark.asyncio
    async def test_delete_workflow_credentials(self, db: AsyncSession, random_private_workflow: WorkflowOut) -> None:
        """
        Test for deleting the workflow credentials in CRUD Repository.

        Parameters
        ----------
        db : sqlalchemy.ext.asyncio.AsyncSession.
            Async database session to perform query on.
        random_private_workflow : app.schemas.workflow.WorkflowOut
            Random private workflow for testing.
        """
        await CRUDWorkflow.update_credentials(db, workflow_id=random_private_workflow.workflow_id, token=None)

        stmt = select(Workflow).where(Workflow._workflow_id == random_private_workflow.workflow_id.bytes)
        workflow = await db.scalar(stmt)
        assert workflow is not None
        assert workflow.credentials_token is None
